﻿// Created by Naoto Ohshima. Date:2007/11/30 23:13:32
using System;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using OhymosNet;

namespace OhymosNet
{
    /// <summary>
    /// OHyMoS.NETが要素モデルと動的にリンクする機能を提供するクラスです。
    /// </summary>
    /// <remarks>
    /// 具体的には、
    /// .NETのリフレクション機能を提供するクラスAssemblyを用いて、
    /// ディクショナリ
    /// <a href = "OhymosNet.OhAssembly._elementAssemblyDic.html">_elementAssemblyDic</a>に、
    /// OhymosNet.dllと
    /// ユーザ環境変数 OHYMOS_LIB に設定されたディレクトリに格納されている
    /// アセンブリとその中に含まれるクラスを
    /// <a href ="OhymosNet.OhAssembly.registerAssembly.html">registerAssembly()</a>
    /// を用いて登録します。
    /// そして、_elementAssemblyDic内を検索し、
    /// 構造定義ファイルから与えられるクラス名と一致するクラスを見つけ出せれば
    /// そのクラス名から
    /// <ul>
    /// <li>"<a href = "OhymosNet.OhElement.create.html">OhElement.create(Type)</a></li>
    /// <li>"<a href = "OhymosNet.OhPort.create.html">OhPort.create(Type)</a></li>
    /// </ul>
    /// を用いてオブジェクトを生成します。
    /// <p/>
    /// また、OHyMoS.NETに含まれる要素モデル（ユーザの用意するものも含む）の
    /// 一覧を表示する機能も提供します。
    /// </remarks>
    /// <example>
    /// 例えば、下のように利用します。利用の際はOhAssemblyを用いるだけで、
    /// クラスは登録されます。
    /// <code>
    /// string className = "SampleClass";
    /// Assembly a;
    /// 
    /// // OhAssemblyのオブジェクトを生成
    /// OhAssembly ohymosAssembly = OhAssembly.Default;
    /// 
    /// // 登録したディクショナリ内で、与えられたクラス名と一致するクラスを検索
    /// if (ohymosAssembly.ElementAssemblyDic.TryGetValue("OhymosNet." + className, out a))
    /// {
    ///     // 見つかれば型を取得し、オブジェクトを生成
    ///     _elements[id] = OhElement.create(a.GetType("OhymosNet." + className, true));
    ///     _elements[id].init(objectName, objectNumber, sr);
    /// }
    /// else
    ///     // 見つからなければ、例外をスロー
    ///     throw new TypeLoadException("\"" + className + "\"" + " does not exist in OHyMoS.NET.");
    /// </code>
    /// </example>
    public class OhAssembly
    {
        /// <summary>
        /// OhAssemblyのシングルトンです。
        /// </summary>
        internal static OhAssembly instance;

        /// <summary>
        /// OHyMoS.NETに用意されている抽象クラスとその各クラスを
        /// 提供しているアセンブリの組を格納したディクショナリです。
        /// </summary>
        internal Dictionary<string, Assembly> _abstractClassAssemblyDic;

        /// <summary>
        /// OHyMoS.NETに用意されている要素モデルとその各モデルを
        /// 提供しているアセンブリの組を格納したディクショナリです。
        /// 
        /// </summary>
        internal Dictionary<string, Assembly> _elementAssemblyDic;

        /// <summary>
        /// OHyMoS.NETに用意されている端子モデルとその各モデルを
        /// 提供しているアセンブリの組を格納したディクショナリです。
        /// 
        /// </summary>
        internal Dictionary<string, Assembly> _portAssemblyDic;

        /// <summary>
        /// OHyMoS.NETに用意されている全てのクラスとその各クラスを
        /// 提供しているアセンブリの組を格納したディクショナリです。
        /// 
        /// </summary>
        internal Dictionary<string, Assembly> _classAssemblyDic;
       
    　　/// <summary>
        /// OhAssemblyの新しいオブジェクトを生成します。
        /// 生成した段階でOHyMoS.NETの全ての要素モデル、端子モデルが
        /// 登録されます。
        /// </summary>
        public OhAssembly()
        {
            _abstractClassAssemblyDic = new Dictionary<string, Assembly>();
            _elementAssemblyDic = new Dictionary<string, Assembly>();
            _portAssemblyDic = new Dictionary<string, Assembly>();
            _classAssemblyDic = new Dictionary<string, Assembly>();

            registerAssembly();
        }

        /// <summary>
        /// OHyMoS.NETに用意されている抽象クラスとその各クラスを
        /// 提供しているアセンブリの組を格納したディクショナリを
        /// 取得します。
        /// 
        /// </summary>
        public Dictionary<string, Assembly> AbstractClassAssemblyDic
        {
            get { return _abstractClassAssemblyDic; }
        }

        /// <summary>
        /// OHyMoS.NETに用意されている要素モデルとその各モデルを
        /// 提供しているアセンブリの組を格納したディクショナリを
        /// 取得します。
        /// 
        /// </summary>
        public Dictionary<string, Assembly> ElementAssemblyDic
        {
            get { return _elementAssemblyDic; }
        }
        /// <summary>
        /// OHyMoS.NETに用意されている端子モデルとその各モデルを
        /// 提供しているアセンブリの組を格納したディクショナリを
        /// 取得します。
        /// 
        /// </summary>
        public Dictionary<string, Assembly> PortAssemblyDic
        {
            get { return _portAssemblyDic; }
        }

        /// <summary>
        /// OHyMoS.NETに用意されている全てのクラスとその各クラスを
        /// 提供しているアセンブリの組を格納したディクショナリを
        /// 取得します。
        /// 
        /// </summary>
        public Dictionary<string, Assembly> ClassAssemblyDic
        {
            get { return _classAssemblyDic; }
        }

        /// <summary>
        /// OhAssemblyのシングルトンを取得します。
        /// </summary>
        public static OhAssembly Default
        {
            get 
            { 
                if(instance==null)
                    instance = new OhAssembly();
                
                return instance;
            }
        }

        /// <summary>
        /// OHyMoS.NETのクラスライブラリOhymosNet.dll内のクラスを登録します。
        /// </summary>
        /// <remarks>
        /// OhymosNet.dll内のクラスを、
        /// <ul>
        /// <li>抽象クラス　->  _abstractClassAssemblyDic</li>
        /// <li>要素モデル  ->  _elementAssemblyDic</li>
        /// <li>端子モデル　->  _portAssemblyDic</li>
        /// <li>全クラス　　->  _classAssemblyDic</li>
        /// </ul>
        /// に登録します。<p/>
        /// このメソッドは<a href ="OhymosNet.OhAssembly.registerAssembly.html">
        /// registerAssembly</a>で使用されます。
        /// </remarks>
        public void makeInitialDic()
        {
            Assembly a = Assembly.GetExecutingAssembly();
            Type[] tArray = a.GetTypes();
            foreach(Type t in tArray)
            {
                if(t.IsAbstract)
                    _abstractClassAssemblyDic.Add(t.FullName,a);

                if(!t.IsAbstract && t.IsSubclassOf(typeof(OhElement)))
                    _elementAssemblyDic.Add(t.FullName, a);

                if(!t.IsAbstract && t.IsSubclassOf(typeof(OhPort)))
                    _portAssemblyDic.Add(t.FullName, a);

                _classAssemblyDic.Add(t.FullName, a);
            }
        }

        /// <summary>
        /// OHyMoS.NET内の全てのクラスを登録します。
        /// </summary>
        /// <remarks>
        /// まず、<a href ="OhymosNet.OhAssembly.makeInitialDic.html">makeInitialDic()</a>
        /// により、OhymosNet.dll内のクラスを登録します。<p/>
        /// 続いて、OHYMOS_LIBに設定されたディレクトリ内に格納されたdllファイル内のクラスを
        /// 登録します。
        /// </remarks>
        public void registerAssembly()
        {
            makeInitialDic();
            
            try
            {
                // OHYMOS_LIBに設定された文字列を取得
                string libDir = Environment.GetEnvironmentVariable("OHYMOS_LIB", EnvironmentVariableTarget.User);
                if (libDir == null)
                    return;

                string[] split = new string[1] { ";" };
                
                // libDirをディレクトリパスに分割
                string[] libDirArray = libDir.Split(split, StringSplitOptions.RemoveEmptyEntries);
                
                foreach (string dir in libDirArray)
                {
                    string[] libArray = Directory.GetFiles(dir, "*.dll");

                    foreach (string str in libArray)
                    {
                        Assembly a = Assembly.LoadFrom(str);
                        Type[] tArray = a.GetTypes();
                        foreach (Type t in tArray)
                        {
                            if (!t.IsAbstract && t.IsSubclassOf(typeof(OhElement)))
                                _elementAssemblyDic.Add(t.FullName, a);

                            else if (!t.IsAbstract && t.IsSubclassOf(typeof(OhPort)))
                                _portAssemblyDic.Add(t.FullName, a);

                            _classAssemblyDic.Add(t.FullName, a);
                        }
                    }
                }
            }

            catch (ArgumentNullException e)
            {
                // OHYMOS_LIBにdllファイルがない場合、makeInitialDic()だけ
                // 実行してreturn
                return;
            }

        }

        /// <summary>
        /// 引数aDicを50音順にソートして表示します。
        /// </summary>
        /// <param name="aDic">a <var>Dictionary &lt; string, ssembly &gt; </var> value<p/>
        /// </param>
        public void listClasses(Dictionary<string, Assembly> aDic)
        {
            SortedDictionary<string, Assembly> sdic = new SortedDictionary<string, Assembly>(aDic);
            Regex regex1 = new Regex(@"^(OhymosNet\.)(\w+)");
            Regex regex2 = new Regex(@"^(.+\\)(\w+\.\w+)$");
                
            foreach (KeyValuePair<string, Assembly> kvp in sdic)
            {
                string str1 = kvp.Key;
                string str2 = kvp.Value.Location;
                Match m1 = regex1.Match(str1);
                Match m2 = regex2.Match(str2);
                if (m1.Success && m2.Success)
                {
                    int[] tabpos = new int[]{0,30};
                    OhAllignText line = new OhAllignText(tabpos);
                    line.clear();
                    line.put(m1.Groups[2].ToString());
                    line.nextColumn();
                    line.put(": " + m2.Groups[2].ToString());
                        Console.WriteLine(line.ToString());
                }
            }
        }

        /// <summary>
        /// 50音順にOHyMoS.NET内の抽象クラスを表示します。
        /// </summary>
        public void listAbstractClasses()
        {
            listClasses(_abstractClassAssemblyDic);
        }

        /// <summary>
        /// 50音順にOHyMoS.NET内の要素モデルを表示します。
        /// </summary>
        public void listElements()
        {
            listClasses(_elementAssemblyDic);
        }

        /// <summary>
        /// 50音順にOHyMoS.NET内の端子モデルを表示します。
        /// </summary>
        public void listPorts()
        {
            listClasses(_portAssemblyDic);
        }

        /// <summary>
        /// 50音順にOHyMoS.NET内の全クラスを表示します。
        /// </summary>
        public void listAllClasses()
        {
            listClasses(_classAssemblyDic);
        }
            
    }
}
     