﻿/// <(#)>  OhInputDB; abstract class to express port models of input files. </(#)>
/// Modified and tested by Naoto Ohshima. Date:2008/01/22 14:14:58
/// Last update:2008/02/26 14:51:58
using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Data;
using System.Collections;
using OhymosNet;

namespace OhymosNet
{

    /// <summary> 
    /// データベース用入力端子モデルを表すクラスです。
    /// このクラスは<a href="OhymosNet.OhInputPort.html">OhInputPort</a>を継承します。<p/>
    /// The OhInputDB class is an abstract class of port models of
    /// input files. It extends 
    /// <a href="OhymosNet.OhInputPort.html">OhInputPort</a> class.<p/>
    /// </summary>
    /// <remarks>
    /// データベースによる入力のための機能を提供するクラスです。
    /// データベースにはMicrosoft社が提供しているSQLServer、フリーソフトウェアの
    /// PostgreSQL、Oracle社が提供しているOracle Databaseなど様々なものがあります。<p/>
    /// 
    /// したがって、OhInputDBにはデータベースに共通な性質として
    /// <ul>
    /// <li>サーバ名（<a href ="OhymosNet.OhInputDB._serverName.html">_serverName</a>）</li>
    /// <li>データベース名（<a href ="OhymosNet.OhInputDB._databaseName.html">_databaseName</a>）</li>
    /// </ul>
    /// また、シミュレーションに用いる列として
    /// <ul>
    /// <li>時刻に関する列の名前（<a href ="OhymosNet.OhInputDB._timeColumns.html">_timeColumns</a>）</li>
    /// <li>データ値に関する列の名前（<a href ="OhymosNet.OhInputDB._valueColumn.html">_valueColumn</a>）</li>
    /// </ul>
    /// をメンバ変数として定義しています。なお、テーブル名についてはオブジェクト名
    /// （<a href ="OhymosNet.OhObject._objectName.html">OhObject._objectName</a>）をテーブル名と
    /// して定義することにします。
    /// <p/> 
    /// <br/>
    /// 基本的にここで定義されるメンバ変数はデータベースファイルに記述され、
    /// 実行時に読み取られることになります。<p/>
    /// データベースファイルとは
    /// <ul>
    /// <li>サーバ名（<a href ="OhymosNet.OhInputDB._serverName.html">_serverName</a>）</li>
    /// <li>データベース名（<a href ="OhymosNet.OhInputDB._databaseName.html">_databaseName</a>）</li>
    /// <li>時刻に関する列の名前（<a href ="OhymosNet.OhInputDB._timeColumns.html">_timeColumns</a>）</li>
    /// <li>データ値に関する列の名前（<a href ="OhymosNet.OhInputDB._valueColumn.html">_valueColumn</a>）</li>
    /// </ul>
    /// を１行ごとに記述したテキストファイルのことです。
    /// 構造定義ファイルの入力端子に関するセクション（[InputPorts]）で
    /// <a href="OhymosNet.OhInputDB.html">OhInputDB</a>の派生クラスである
    /// 各データベース、データ型に対応した実装クラスを指定し、データベースファイルを
    /// 以下のように記述することで、データベースを利用することができます。<p/>
    /// <gt>例：</gt>
    /// <code>
    /// # [InputPorts]
    /// # 通し番号　テーブル名　クラス名　データベースファイル名
    /// 0 table0 OhInputDBSqlDouble total.db 
    /// </code>
    /// </remarks>
    public abstract class OhInputDB : OhInputPort
    {
        // @begin {instance variable or static variable} -------------

       /// <summary>
        /// データベースのあるサーバ名です。
        /// </summary>
        protected internal string _serverName;

        /// <summary>
        /// データベース名です。
        /// </summary>
        protected internal string _databaseName;

        /// <summary>
        /// シミュレーションに使用する時系列の列名を格納する配列です。
        /// </summary>
        /// <remarks>
        /// 使用するテーブルの列から、シミュレーションに使用する時系列
        /// （年、月、日など）の列名を格納する文字列配列です。
        /// 例えば、
        /// 使用する列名がyear、month、day、hourのときは
        /// <code>string[] _timeColumns = {year, month, day, hour}</code>
        /// となります。
        /// 
        /// </remarks>
        protected internal string[] _timeColumns;

        /// <summary>
        /// シミュレーションに使用するテーブルの時系列の種類を表します。<p/>
        /// 種類の詳細については<a href = "OhymosNet.OhInputDB.InputDateType.html">
        /// InputDateType</a>を参照してください。
        /// </summary>
        protected internal InputDateType _inputDateType;

        /// <summary>
        /// シミュレーションに使用するデータ値の列名です。
        /// </summary>
        protected internal string _valueColumn;

        /// <summary>
        /// データベースファイル名です。
        /// </summary>
        protected internal string _dbFile;
        // @end {instance variable or static variable} ---------------


        // @begin {abstract property and methods} ---------------------------------

        // 1. connection function

        /// <summary> 
        /// 送信デ－タパック型を表す文字列を取得します。<p/>
        /// Gets the string showing the type of the sending datapacks.
        /// <p/>
        /// </summary>
        public abstract override string DataType { get; }

        // 2. sending function of the input file data

        /// <summary> 
        /// テーブルの行データからデータパック作成を作成します。<p/>
        /// Makes datapacks from a row data in data table.
        /// </summary>
        /// <remarks>
        /// テーブルの行データからデータパック作成を作成します。<p/>
        /// 抽象メソッドです。
        /// 派生クラスで実装することが要求されています。<p/>
        /// 
        /// １行分の行データを１個のデータパックに変換するメソッドを
        /// 派生クラスで定義して下さい。<p/>
        /// 
        /// 標準的な派生クラスでは、
        /// 送信データパック型に応じて次の手続きを定義して下さい。<p/>
        /// 
        /// <ul>
        /// <li>(1) 送信用データパックの記憶領域をヒープに生成する</li>
        /// <li>(2) 行データ aDr からデータ記録時刻、データ値を読み取る。</li>
        /// <li>(3) (2) で読み取った値を (1) のデータパックのメンバに代入する。</li>
        /// <li>(4) (1) のデータパックを OhDataPack 型にキャストして返す。</li>
        /// </ul>
        /// 
        /// このクラスで用意されているメソッド
        /// <a href="OhymosNet.OhInputDB.convertDateTime.html">convertDateTime(DataRow)</a>
        /// を用いると、
        /// <code>
        /// public override OhDataPack makeDataPack(DataRow aDr)
        /// {
        ///     IOhTimePointImmutable dateTime = convertDateTime(aDr);
        ///     double value = (double)aDr[_valueColumn];
        ///     OhDataPackDouble dp = new OhDataPackDouble(dateTime,value);
        ///     return (OhDataPack)dp;
        /// }
        /// </code>
        /// のように書けます。
        /// 
        /// 抽象メソッドなので派生クラスで必ず定義して下さい。
        /// </remarks>
        /// 
        /// <param name="aDr">a <var>DataRow</var> value<p/>
        /// データテーブルの行データ
        /// </param>
        /// <returns> an <var>OhDataPack</var> value<p/>
        /// 送信データパックを OhDataPack 型にキャストしたもの
        /// </returns>
        public abstract OhDataPack makeDataPack(DataRow aDr);

        /// <summary>
        /// データセットを作成します。<p/>
        /// Makes data set.
        /// </summary>
        /// <remarks>
        /// データベースに接続し、データセットを作成します。
        /// 抽象メソッドです。
        /// <p/>
        /// データセットとは、データベースから取得したデータをメモリ上に
        /// 保持しておくための仕組みです。
        /// 各種類のデータベースに対応したデータプロバイダを用いなければならないため、
        /// 派生クラスで実装する必要があります。
        /// 例えば、SQLServerを用いる場合はmakeDataSet()は以下のように記述します。<p/>
        /// 例:
        /// <code>
        /// 
        /// </code>
        /// </remarks>
        /// <returns></returns>
        public abstract DataSet makeDataSet();

        // @end {abstract methods} -----------------------------------


        // @begin{initialization} ------------------------------------

        /// <summary> 
        /// オブジェクト名(テーブル名)を aFileName、オブジェクト番号を 0 に初期化します。<p/>
        /// Initialization.
        /// </summary>
        /// <param name="aTableName">a <var>String</var> value<p/>
        /// テーブル名
        /// </param>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public virtual void init(String aTableName)
        {
            base.init(aTableName, 0);
        }

        /// <summary>
        /// オブジェクト名をaTableName、オブジェクト番号をaTableNumに初期化し、
        /// データベースファイルaDbFileを読み取ります。
        /// </summary>
        /// <remarks>
        /// データベースファイルとは
        /// <ul>
        /// <li>サーバ名（<a href ="OhymosNet.OhInputDB._serverName.html">_serverName</a>）</li>
        /// <li>データベース名（<a href ="OhymosNet.OhInputDB._databaseName.html">_databaseName</a>）</li>
        /// <li>時刻に関する列の名前（<a href ="OhymosNet.OhInputDB._timeColumns.html">_timeColumns</a>）</li>
        /// <li>データ値に関する列の名前（<a href ="OhymosNet.OhInputDB._valueColumn.html">_valueColumn</a>）</li>
        /// </ul>
        /// を１行ごとに記述したテキストファイルのことです。<p/>
        /// 
        /// このメソッドは構造定義ファイルを用いる全体系モデル
        /// <a href = "OhymosNet.OhScfTotalSystem.html">OhScfTotalSystem</a>のコンストラクタ内で
        /// 使用されます。
        /// 
        /// </remarks>
        /// <param name="aTableName">テーブル名</param>
        /// <param name="aTableNum">オブジェクト番号</param>
        /// <param name="aDbFile">データベースファイル名</param>
        public override void init(string aTableName, int aTableNum, OhStreamReader aDbFile)
        {
            base.init(aTableName, aTableNum);

            if(_dbFile==null) // １回目だけ読み込む（dbファイルの記述間違い等で複数回init()を呼び出すことがある）
                _dbFile = aDbFile.readWord();
            
            OhStreamReader dbSr = new OhStreamReader(_dbFile, true);
            
            try
            {
                _serverName = dbSr.readWord();
                dbSr.skipWhites();
                _databaseName = dbSr.readWord();
                dbSr.skipWhites();
                string timeLine = dbSr.ReadLine();
                string[] split ={ " " };
                _timeColumns = timeLine.Split(split, StringSplitOptions.RemoveEmptyEntries);
                _inputDateType = (InputDateType)_timeColumns.Length;
                _valueColumn = dbSr.readWord();
                //sr.skipWhites();

                // できればここで時間範囲指定
            }

            catch (IOException e)
            {
                throw new OhError("error occured.", oPrintString("\t"));
            }
            finally
            {
                dbSr.Close();
            }
        }
        
        // @end{initialization} --------------------------------------


        // @begin {constructors} -------------------------------------

        /// <summary> 
        /// デフォルトコンストラクタです。
        /// Creates a new default <var>OhInputDB</var> instance.
        /// 
        /// </summary>
        /// <remarks>デフォルトコンストラクタは
        /// <a href="OhInputDB.html#init(String, int)">OhInputDB.init(String, int)</a> を実行しません。
        /// その結果、以下の設定となります。<p/>
        /// 
        /// <ul>
        /// <li> _objectName        : null</li>
        /// <li> _objectNumber      : -1</li>
        /// <li> _className         : "OhInputDB"</li>
        /// <li> _fileTimeAtInput   : Long.MinValue (考え得る最小 long 値)</li>
        /// </ul>
        /// 
        /// 
        /// <p/>
        /// その他のメンバ変数に対する処理は引数つきコンストラクタと同じです。
        /// この場合、<a href="OhInputDB.html#init(String, int)">OhInputDB.init(String, int)</a> を
        /// 用いてメンバ変数を初期化して下さい。
        /// </remarks>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public OhInputDB()
            : base()
        {
            _className = "OhInputDB";
        }

        /// <summary> 
        /// オブジェクト（ファイル）名を指定して <var>OhInputDB</var> クラスの
        /// インスタンスを生成します。<p/>
        /// Creates a new <var>OhInputDB</var> instance.
        /// </summary>
        /// <remarks>
        /// 基底クラス <a href="OhInputPort.html">OhInputPort</a> のデフォルトコンストラクタを用い、
        /// <a href="OhInputDB.html#init(String, int)">OhInputDB.init(String, int)</a> を用いて、
        /// メンバ変数を以下の通り初期化します。<p/>
        /// 
        /// <dl>
        /// <li> _className         : "OhInputDB"</li>
        /// <li> _objectName        : aTableName</li>
        /// <li> _objectNumber      : 0</li>
        /// <li> _owner             : null</li>
        /// <li> _bufferList        : メモリに確保する。(接続作業用)</li>
        /// <li> _receiverArray     : null</li>
        /// <li> _numberOfReceivers : -1</li>
        /// <li> _dataPackArray     : null</li>
        /// <li> _numberOfDataPacks : -1</li>
        /// </dl>
        /// </remarks>
        /// <param name="aTableName">a <var>String</var> value<p/>
        /// ファイル名
        /// </param>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public OhInputDB(String aTableName)
            : base()
        {
            _className = "OhInputDB";
            _objectName = aTableName;
            _objectNumber = 0;
        }

        // @end {constructors} ---------------------------------------

        // @begin {methods} ------------------------------------------

        /// <summary> 
        /// 端子は計算機能がないので、計算要請があってもどうしょうもない。
        /// 必ず例外 <a href="OhymosNet.OhCannotGetDataError.html">OhCannotGetDataError</a> を投げます。
        /// <p/>
        /// When the input port receives a calculation request, throw
        /// <a href="OhymosNet.OhCannotGetDataError.html">OhCannotGetDataError</a>. 
        /// </summary>
        /// <remarks>
        /// ここでスローされた例外は<a href = "OhymosNet.OhTotalSystem.work.html">
        /// OhTotalSystem.work()</a>でキャッチされます。<p/>
        /// This will be caught in the catch
        /// block in OhTotalSystem.work().
        /// </remarks>
        /// <exception cref="OhCannotGetDataError">常に例外をスローします
        /// </exception>
        public override void calReq()
        {
            String pauseComment = "Cannot get data from " + _objectName;

            OhInterface.pause(pauseComment);
            throw new OhCannotGetDataError(pauseComment);
        }

        /// <summary> 
        /// この入力ファイルの情報をストリーム <b>aSw</b> に出力します。
        /// 出力するときに、各行の行頭に、文字列 <b>aHead</b> を付加し
        /// ます。<p/>
        /// <p/>
        /// 
        /// The information on this input file is printed to the
        /// specified PrintWriter <b>aSw</b> with the preceding String
        /// <b>aHead</b>
        /// </summary>
        /// <param name="aSw">a <var>StreamWriter</var> value<p/>
        /// オブジェクト情報を書き出すストリーム
        /// </param>
        /// <param name="aHead">a <var>String</var> value<p/>
        /// 各行頭に付加する文字列
        /// </param>
        public override void cPrint(StreamWriter aSw, String aHead)
        {
            aSw.WriteLine(aHead + _className + " " + _objectName);
        }

        /// <summary>
        /// テーブルに格納されたデータのうち、dbファイル内で指定した列を
        /// 最初の５行まで表示します。
        /// OhInputDB.inputData()の補助メソッドです。
        /// </summary>
        /// <param name="aTable"></param>
        public void showDataTable(DataTable aTable)
        {
            Console.WriteLine("テーブル {0} は以下の通りです（最初の5行）", aTable.TableName);

            DataColumnCollection columns = aTable.Columns;
            DataRowCollection rows = aTable.Rows;

            foreach (DataColumn dc in columns)
                Console.Write(dc.ColumnName + "\t");
            Console.WriteLine();

            StringBuilder sb = new StringBuilder();
            foreach (DataColumn dc1 in columns)
                sb.Append(dc1.ToString());

            int columnsLength = sb.ToString().Length + columns.Count * 4;
            for (int i = 0; i < columnsLength; i++)
                Console.Write("-");
            Console.WriteLine();

            int count=5;
            if(rows.Count<5)
                count=rows.Count;

            for (int k = 0; k < count; k++ )
            {
                for (int i = 0; i < columns.Count; i++)
                {
                    Console.Write(rows[k][i] + "\t");
                }
                Console.WriteLine();
            }
        }

        /// <summary> 
        /// データベースに格納された入力テーブルを読み取り、データパックを作成し、送信します。
        /// </summary>
        /// <remarks>
        /// 入力テーブルを読み取り、
        /// <a href="OhInputFile.html#makeDataPack(String)">OhInputDB.makeDataPack(DataRow)</a> により
        /// データパックを作成し、作成したデータパックを
        /// <a href="OhSendPort.html#dataPackStock(Object)">OhSendPort.dataPackStock(OhDataPack)</a> を
        /// 用いて送信します(OhInputDB は OhSendPort の一種です)。<p/>
        /// </remarks>
        public override void inputData()
        {
            setDataPackArray();

            reduceDataPackStock(); // Delete the unnecessary datapacks.


            DataSet ds = makeDataSet(); // データセットの作成

            // オブジェクト名で指定されたテーブル名のDataTableオブジェクトの生成
            DataTable dt = ds.Tables[_objectName]; 

            // テーブルの表示
            showDataTable(dt);

            // データパックの生成
            foreach (DataRow dr in dt.Rows)
            {
                dataPackStock(makeDataPack(dr));
            }
                
            setDataPackArray(); //; Set "_numberOfDataPacks = _dataPackArray.size();"

            if (NumberOfDataPacks > 0)
            {
                IOhTimePointImmutable tp = NewestDataTime;

                StringWriter sw = new StringWriter();
                sw.WriteLine("Server: "+ _serverName + " Database: "+_databaseName + NEWLINE 
                    +"Table: " + _objectName + "  newest data time : "
                    + tp.Year + " " + tp.Tsec + " " + tp.Frac);

                String str = sw.ToString();

                OhObject top = (OhObject)TopOwner;
                if (top != null && top is OhTotalSystem)
                {
                    // (Ref "OhTotalSystem#addNewestDataTime()")
                    ((OhTotalSystem)top).addNewestDataTime(str);
                }
            }
        }

       

        /// <summary> 
        /// この入力ファイルの情報をストリーム <b>aSw</b> に出力します。
        /// 出力するときに、各行の行頭に、文字列 <b>aHead</b> を付加し
        /// ます。
        /// <p/>
        /// 
        /// The information on this input file is printed to the
        /// specified PrintWriter <b>aSw</b> with preceding String
        /// <b>aHead</b>
        /// </summary>
        /// <param name="aSw">a <var>StreamWriter</var> value<p/>
        /// 入力ファイルの情報を書き込むストリーム
        /// </param>
        /// <param name="aHead">a <var>String</var> value<p/>
        /// </param>
        public override void oPrint(StreamWriter aSw, String aHead)
        {
            cPrint(aSw, aHead);
        }

        /// <summary>
        /// シミュレーションに使用するテーブルの時系列の種類を表します。
        /// </summary>
        /// <remarks>
        /// 例えば、時刻の表現として
        /// <ul>
        /// <li>年、月、日、時、分、秒</li>
        /// <li>年、月、日、時</li>
        /// <li>年、年始からの秒</li>
        /// </ul>
        /// などの違いを表現するものです。<p/>
        /// この列挙体は<a href = "OhymosNet.OhInputDB.convertDateTime.html">convertDateTime(DataRow)</a>内で
        /// これらの時刻表現とデータ値から<a href = "OhymosNet.OhTimePoint.html">OhTimePoint</a>型に変換する
        /// 際に使用します。<p/>
        /// なお、OHyMoS.NETで用いられる時刻は必ず西暦年を指定する必要があります。
        /// </remarks>
        public enum InputDateType
        {
            /// <summary>
            /// テーブル内の時刻表現が年、月、日、時、分、秒となっていることを表します。
            /// </summary>
            SEC=6,

            /// <summary>
            /// テーブル内の時刻表現が年、月、日、時、分となっていることを表します。
            /// </summary>
            MIN=5,

            /// <summary>
            /// テーブル内の時刻表現が年、月、日、時となっていることを表します。
            /// </summary>
            HOUR=4,

            /// <summary>
            /// テーブル内の時刻表現が年、月、日となっていることを表します。
            /// </summary>
            DAY=3,
            //MONTH=,
            //YEAR,
            /// <summary>
            /// テーブル内の時刻表現が年、年始からの秒となっていることを表します。
            /// </summary>
            TSEC = 2,
        }

        //  OhTimePointの静的メソッドにした方がいいかも↓

        /// <summary>
        /// シミュレーションに用いる時刻に関する列の行データを
        /// <a href = "OhymosNet.IOhTimePointImmutable.html">IOhTimePointImmutable</a>型に
        /// 変換します。
        /// </summary>
        /// <remarks>
        /// このメソッドはデータベースに格納されている任意の時刻表現のデータを
        /// IOhTimePointImmutable型に変換します。<p/>
        /// OhInputDBSqlDoubleなどの
        /// <a href ="OhymosNet.OhInputDBSqlDouble.makeDataPack.html">makeDataPack(DataRow)</a>で
        /// 使用されます。
        /// </remarks>
        /// <param name="aDr">変換する行</param>
        /// <returns>a <var>IOhTimePointImmutable</var> value<p/>
        /// </returns>
        public IOhTimePointImmutable convertDateTime(DataRow aDr)
        {
            int year = 0;
            int month = 0;
            int day = 0;
            int hour = 0;
            int min = 0;
            int sec = 0;
            long tsec = 0;
            IOhTimePointImmutable dateTime;

            switch (_inputDateType)
            {
                case InputDateType.SEC:
                    year = (int)aDr[0];
                    month = (int)aDr[1];
                    day = (int)aDr[2];
                    hour = (int)aDr[3];
                    min = (int)aDr[4];
                    sec = (int)aDr[5];
                    dateTime = new OhTimePoint(year, month, day, hour, min, sec);
                    break;

                case InputDateType.MIN:
                    year = (int)aDr[0];
                    month = (int)aDr[1];
                    day = (int)aDr[2];
                    hour = (int)aDr[3];
                    min = (int)aDr[4];
                    dateTime = new OhTimePoint(year, month, day, hour, min, sec);
                    break;

                case InputDateType.HOUR:
                    year = (int)aDr[0];
                    month = (int)aDr[1];
                    day = (int)aDr[2];
                    hour = (int)aDr[3];
                    dateTime = new OhTimePoint(year, month, day, hour, min, sec);
                    break;

                case InputDateType.DAY:
                    year = (int)aDr[0];
                    month = (int)aDr[1];
                    day = (int)aDr[2];
                    dateTime = new OhTimePoint(year, month, day, hour, min, sec);
                    break;

                case InputDateType.TSEC:
                    year = (int)aDr[0];
                    tsec = (long)aDr[1];
                    dateTime = new OhTimePoint(year, tsec);
                    break;

                default:
                    throw new OhError("Undefined format is used.");
            }
            return (IOhTimePointImmutable)dateTime;
        }

        /// <summary>
        /// このメソッドはOhInputDBでは使用されません。
        /// 必ず例外を投げます。
        /// </summary>
        /// <param name="aDstr">データ文字列</param>
        public override void sendByString(string aDstr)
        {
            throw new OhError("This method is not implemented in OhInputDB.");
        }
    
       
        // @end {methods} --------------------------------------------
    } // OhInputDB
}