﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using Npgsql;

namespace OhymosNet
{
    /// <summary>
    /// PostgreSQL用入力端子モデルを表すクラスです。
    /// 抽象クラスです。
    /// このクラスは<a href="OhymosNet.OhInputDB.html">OhInputDB</a>を継承しています。
    /// <p/>
    /// The OhInputDBPgSql class is an abstract class of port models of
    /// input database for PostgreSQL. It extends 
    /// <a href="OhymosNet.OhInputDB.html">OhInputDB</a> class.<p/>
    /// </summary>
    /// <remarks>
    /// PostgreSQL用の入力端子モデルを表すクラスです。
    /// このクラスではデータベースファイルから読み取った情報と
    /// コマンドラインから入力されるユーザID、パスワードを基に、
    /// SQLServerデータベースに接続します。<p/>
    /// そして、データベースから取得したデータを保持する
    /// データセットオブジェクトを作成します。
    /// 
    /// </remarks>
    public abstract class OhInputDBPgSql : OhInputDB
    {
        /// <summary>
        /// デフォルトコンストラクタです。
        /// </summary>
        /// <remarks>
        /// デフォルトコンストラクタは
        /// <a href="OhymosNet.OhInputDBPgSql.init_overload_1.html">OhInputDBPgSql.init(String, int)</a> 
        /// を実行しません。その結果、以下の設定となります。<p/>
        /// <dl>
        /// <li> _objectName        : null</li>
        /// <li> _objectNumber      : -1</li>
        /// </dl>
        /// 
        /// <p/>
        /// デフォルトコンストラクタを使用する場合、
        /// <a href="OhymosNet.OhInputDBPgSql.init_overload_1.html">OhInputDBPgSql.init(String, int)</a> 
        /// を用いてメンバ変数を初期化して下さい。
        /// </remarks>
        public OhInputDBPgSql()
            : base()
        {
            _className = "OhInputDBPgSql";
        }

        /// <summary>
        /// 入力端子名、入力端子番号を初期化します。
        /// </summary>
        /// <param name="aPortName">入力端子名</param>
        /// <param name="aPortNum">入力端子番号</param>
        public override void init(string aPortName, int aPortNum)
        {
            base.init(aPortName, aPortNum);
        }

        /// <summary>
        /// データベースに接続して、データセットを作成します。
        /// </summary>
        /// <remarks>
        /// まず、ユーザーIDとパスワードの
        /// 入力を求められるので入力してください。
        /// IDとパスワード（とdbファイルの内容）が正しければデータベースに接続します。
        /// 正しくない場合はエラーメッセージを出力します。<p/>
        /// 
        /// その後、dbファイルで指定された列
        /// （_timeColumns、_valueColumn）
        /// だけを選択し、
        /// その列で構成されるデータセットを作成します。
        /// </remarks>
        /// <returns>
        /// a <var>DataSet</var> value<p/>
        /// 作成したデータセット
        /// </returns>
        /// <exception cref="OhError">
        /// ユーザID、パスワード、またはdbファイルの内容が間違っている場合
        /// </exception>
        public override DataSet makeDataSet()
        {
            Console.WriteLine("サーバ\t\t{0}", _serverName);
            Console.WriteLine("データベース\t{0}", _databaseName);
            Console.WriteLine("テーブル\t{0}", _objectName);
            Console.WriteLine("に接続します");
            Console.WriteLine();
            // DataSetのインスタンスの作成
            DataSet ds = new DataSet();
            bool initFlag = false;
            while (true)
            {
                Console.Write("データベースに接続するためのユーザーIDを入力してください" + NEWLINE + ">");
                string userID = Console.ReadLine();
                Console.Write("パスワードを入力してください" + NEWLINE + ">");
                string password = Console.ReadLine();
                
                string objectName;
                int objectNumber;
                OhStreamReader sr;
                if (initFlag)
                {
                    sr = new OhStreamReader(_dbFile);
                    objectName = _objectName;
                    objectNumber = _objectNumber;
                    _objectName = null;
                    _objectNumber = -1;
                    init(objectName, objectNumber, sr);
                }
                // データベース接続のための接続文字列
                string connStr = "Server = " + _serverName
                                + "; User Id = " + userID
                                + "; Password = " + password
                                + "; Database = " + _databaseName;

                // レコード取得のためのクエリ
                StringBuilder sb = new StringBuilder("select ");
                foreach (string str in _timeColumns)
                    sb.Append(str + ",");
                sb.Append(_valueColumn + " from " + _objectName);

                string sqlStr = sb.ToString();

                // SqlDataAdapterのインスタンスの作成
                NpgsqlDataAdapter da = new NpgsqlDataAdapter(sqlStr, connStr);

                try
                {
                    Console.WriteLine("Connecting the database...");

                    // 指定されたクエリを実行し、テーブルをDataSetオブジェクト内に格納
                    da.Fill(ds, _objectName);
                }

                catch (NpgsqlException e)
                {
                    Console.WriteLine("ユーザーID、パスワード、またはdbファイルの内容が間違っています");
                    Console.WriteLine("ユーザーID、パスワード、またはdbファイルの内容を確認してください");
                    initFlag = true;
                    continue;
                }
                break;
            }
            return ds;
        }
    }
}
