﻿/// <(#)>  OhInputFileInt: Derived from OhInputFile. </(#)>
/// Modified and Tested by Naoto Ohshima. Date:2007/12/01 2:05:06
/// <summary> 
/// (Ref "OhObject")
/// <-- (Ref "OhPort")
/// <-- (Ref "OhSenderPort")
/// <-- (Ref "OhSendPort")
/// <-- (Ref "OhInputPort")
/// <-- (Ref "OhInputFile")
/// </summary>
using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using OhymosNet;

namespace OhymosNet
{

    /// <summary> 
    /// ファイルからデータ値が double 型であるデータパックを読み取る入力端子を表すクラスです。
    /// このクラスは<a href ="OhymosNet.OhInputFile.html">OhInputFile</a>を
    /// 継承しています。<p/>
    /// 
    /// The OhInputFileInt class is a class of input port which
    /// reads in datapacks of type double from an input file and
    /// sends them. It extends
    /// <a href ="OhymosNet.OhInputFile.html">OhInputFile</a> class.<p/>
    /// </summary>
    /// <remarks>
    /// 端子を接続するには、接続する各端子のデータ型が合っている必要があります。
    /// したがって、OhInputFileIntは
    /// <ul>
    /// <li><a href="OhymosNet.OhReceiveInt.html">OhReceiveInt</a></li>
    /// <li><a href="OhymosNet.OhRelayInt.html">OhRelayInt</a></li>
    /// <li><a href="OhymosNet.OhOutputFileInt.html">OhOutputFileInt</a></li>
    /// </ul>
    /// と接続することができます。<p/>
    /// 
    /// 読み取るデータは１つのデータパックにつき1行ごとに書かれ、
    /// 下の例のように、スペースまたはタブで区切られているものとします。
    ///  <code>
    /// 2003 120 0.3 1.2 2.3 4.5
    /// </code>
    /// 違う形式で読み取る場合には、このクラスを継承したクラスを作成し、
    /// <a href="OhymosNet.OhInputFileInt.makeDataPack.html">makeDataPack(string)</a>を
    /// 再定義してください。
    /// <br/><br/>
    /// OhInputFileInt can be connected with
    /// <a href="OhymosNet.OhReceiveInt.html">OhReceiveInt</a>,
    /// <a href="OhymosNet.OhRelayInt.html">OhRelayInt</a>, and 
    /// <a href="OhymosNet.OhOutputFileInt.html">OhOutputFileInt</a>.
    /// 
    /// OhInputFileInt reads an input file using "simple data
    /// format for input and output." <p/>
    /// 
    /// "Simple data format for input and output" is a format in
    /// which
    /// 
    /// .....
    /// all the data items for a datapack should be written in one
    /// line, and data items should be separated with a number of
    /// spaces or tabs.
    /// <code>
    /// 2003 120 0.3 1.2 2.3 4.5
    /// </code>
    /// ....
    /// 
    /// If you need different format, please make a derived class of
    /// OhInputFileInt and redefine "<a href="OhymosNet.OhInputFileInt.makeDataPack.html">makeDataPack(string)</a>".
    /// </remarks>
    public class OhInputFileInt : OhInputFile
    {

        /// <summary> 
        /// 入力ファイルから読み取られる値に乗ずる値（乗数）です。<p/>
        /// Multiplication factor to the value which is read in from
        /// the input file. 
        /// </summary>
        /// 
        /// <remarks>
        /// 例えば
        /// <br/><br/>
        /// _mulfactor = 2.0
        /// <br/><br/>
        /// としておくと、入力ファイルから読み取った値が3.2だとすれば
        /// データ値は6.4と設定されます。<p/>
        /// デフォルトでは1.0（入力ファイルから読み取られる値をそのまま使う）としておきます。<p/>
        /// 
        /// For example, assume that _mulfactor is
        /// equal to 2.0 and a numerical value 3.2 is read in from the
        /// input file, then the data value is set to 6.4.
        /// </remarks>
        protected internal double _mulfactor = 1.0; //; 倍率

        // @begin{initialization} ------------------------------------

        /// <summary> 
        /// オブジェクト名(ファイル名)を aFileName、オブジェクト番号を 0 に初期化します。
        /// <p/>
        /// Initializ the object. This method assigns the argument
        /// aFileName to the instance variable
        /// (Ref "OhObject._className").
        /// 
        /// </summary>
        /// <param name="aFileName">a <var>String</var> value<p/>
        /// 入力ファイル名
        /// </param>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public override void init(String aFileName)
        {
            base.init(aFileName);
        }

        // @end{initialization} --------------------------------------


        // @begin {constructors} -------------------------------------

        /// <summary> 
        /// デフォルトコンストラクタです。
        /// <p/>
        /// Creates a new default <var>OhInputFileInt</var>
        /// instance.
        /// </summary>
        /// <remarks>
        /// デフォルトコンストラクタは
        /// <a href="OhInputFileInt.html#init(java.lang.String, int)">OhInputFileInt.init(String, int)</a>
        /// を実行しません。
        /// その結果、以下の設定となります。<p/>
        /// 
        /// <dl>
        /// <li> _objectName        : null</li>
        /// <li> _objectNumber      : -1</li>
        /// </dl>
        /// 
        /// <p/>
        /// その他のメンバ変数に対する処理は引数つきコンストラクタと同じです。
        /// この場合、
        /// <a href="OhInputFileInt.html#init(java.lang.String, int)">OhInputFileInt.init(String, int)</a>
        /// を用いてメンバ変数を初期化して下さい。
        /// </remarks>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public OhInputFileInt()
            : base()
        {
            _className = "OhInputFileInt";
        }


        /// <summary> Creates a new <var>OhInputFileInt</var> instance.
        /// 
        /// </summary>
        /// <remarks>
        /// 基底クラスのコンストラクタ
        /// <a href="OhInputFile.html#OhInputFile(java.lang.String)">OhInputFile.OhInputFile(String)</a>
        /// を用いて、
        /// メンバ変数を以下の通り初期化します。<p/>
        /// 
        /// <dl>
        /// <li> _className         : "OhInputFileInt"</li>
        /// <li> _objectName        : aFileName</li>
        /// <li> _objectNumber      : 0</li>
        /// <li> _owner             : null</li>
        /// <li> _bufferList        : メモリに確保する。(接続作業用)</li>
        /// <li> _receiverArray     : null</li>
        /// <li> _numberOfReceivers : -1</li>
        /// <li> _dataPackArray     : null</li>
        /// <li> _numberOfDataPacks : -1</li>
        /// <li> _fileTimeAtInput   : Long.MIN_VALUE (考え得る最小 long 値)</li>
        /// </dl>
        /// </remarks>
        /// 
        /// <param name="aFileName">a <var>String</var> value<p/>
        /// 入力ファイル名
        /// </param>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public OhInputFileInt(String aFileName)
            : base(aFileName)
        {
            _className = "OhInputFileInt";
        }

        // @end {constructors} ---------------------------------------


        // @begin {property}-----------------------------------------------
        // 1. connection function.

        /// <summary> 
        /// 入力データパックのデータ値の型 int を文字列 "int" を取得します。<p/>
        /// 
        /// Gets the data pack type which expresses the type of data
        /// to be received.
        /// 
        /// </summary>
        public override String DataType
        {
            get { return "int"; }
        }
        // @end {property}-----------------------------------------------

        // @begin {methods} ------------------------------------------

        // 2. data input function

        /// <summary> 
        /// 引数 aLineData に格納されているデータパック文字列からデータ
        /// パックを作成し、OhDataPack 型として戻します。<p/>
        /// 
        /// Reads string datapacks of type int, creates datapacks of
        /// type int, and returns it.
        /// </summary>
        /// 
        /// <remarks>
        /// 以下の例だと、データ時刻 OhTimePoint(2003, 12, 0.3)、にデータ
        /// 値 3.5 を取るデータパックを作成します。
        /// </remarks>
        /// <example>
        /// OhDataPack dataPack = makeDataPack("2003 12 0.3 3.5");<p/>
        /// </example>
        /// 
        /// <param name="aDataStr">a <var>String</var> value<p/>
        /// データパック文字列
        /// </param>
        /// <returns> an <var>OhDataPack</var> value<p/>
        /// </returns>
        public override OhDataPack makeDataPack(String aDataStr)
        {
            OhDataPackInt dp = new OhDataPackInt(aDataStr);
            dp.multiply(_mulfactor);
            return dp;
        }

        /// <summary> 
        /// 引数で与えられた、"#&amp; " で始まる行を解析して、対応して定義
        /// された操作を実行します。
        /// 
        /// The parseSpecLine method analyzes the line which starts
        /// with "#&amp; " and executes the specified operation by the
        /// line.
        /// </summary>
        /// <remarks>
        /// このメソッドの定義は、スーパークラスで、与えられていますが、
        /// このクラスでオーバーライドされています。(参照: (Ref "OhInputFile#partsSpecLine()") ).
        /// <p/>
        /// 
        /// (Ref "OhInputFile#parseSpecLine()") で定義されていたデータ
        /// 補間型配列を読み込む操作に加えて、データ値を定数倍する操作
        /// が定義されています。"#&amp; " の後に multiply と書いた行は、乗
        /// 数が書いてあって、ファイルから読取ったデータ値にこの乗数を
        /// 掛けたものがデータ値として採用されます。<p/>
        /// <br/>
        /// This method is implemented in the super class
        /// (See: (Ref "OhInputFile#parseSpecLine()") ), but is
        /// overrided in this OhInputFileInt class.<p/>
        /// 
        /// In this implementation of the parseSpecLine()
        /// method, the following multiplication specification<p/>
        /// 
        /// <code>
        /// #&amp; multiply 2.0
        /// </code>
        /// 
        /// is alse analyzed. This exmaple line specifies the
        /// multiplication factor to 2.0, that is, the data which are
        /// read in fromt this file are multiplied with 2.0.
        /// <p/>
        /// </remarks>
        /// <param name="aSpecLine">a <var>String</var> value<p/>
        /// 読み取る特別な行
        /// </param>
        /// <exception cref="OhError">if an error occurs
        /// </exception>
        public override void parseSpecLine(String aSpecLine)
        {
            // aSpecLine samples
            // #& dctvec 2 1 0
            // #& multiply 2.0
            Regex regex = new Regex("(\\w+) ");
            Match m = regex.Match(aSpecLine.Substring(3, aSpecLine.Length - 3));
            if (m.Success)
            {
                String keyword = m.Groups[1].ToString();
                String info = aSpecLine.Substring(3 + keyword.Length);
                if (keyword.ToLower().Equals("dctvec"))
                {
                    DctVec = DataConnectionTypeOption.parseDctVec(info);
                }
                else if (keyword.ToLower().Equals("multiply"))
                {
                    _mulfactor = Double.Parse(info);
                }
                else
                {
                    throw new OhError("OhError wad detected." + NEWLINE
                        + " File: " + _objectName + "." + NEWLINE
                        + " in:   " + aSpecLine + NEWLINE
                        + " Unrecognized option: " + keyword + NEWLINE);
                }
            }
            else
            {
                throw new OhError("OhError wad detected." + NEWLINE
                    + " File: " + _objectName + "." + NEWLINE
                    + " Illegal option line: " + aSpecLine + NEWLINE);
            }
        }

        /// <summary> 
        ///  引数 aStr に格納されているデータパック文字列からデータパッ
        /// クを作成し、それを入力端子に収録します。<p/>
        /// 
        /// Stocks the datapack of type int which is stored in the
        /// form of string.
        /// </summary>
        /// <remarks>
        /// メソッドの定義で <a href="OhSendPort.html#dataPackStock()">OhSendPort.dataPackStock()</a> 
        /// を使用しています。<p/>
        /// </remarks>
        /// <param name="aDataStr">a <var>String</var> value<p/>
        /// データパック文字列
        /// </param>
        public override void sendByString(String aDataStr)
        {
            OhDataPackInt dp = new OhDataPackInt(aDataStr);
            dp.multiply(_mulfactor);
            dataPackStock(dp);
        }

        // @end {methods} --------------------------------------------
    } // OhInputFileInt
}