﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace OhymosNet
{
    /// <summary>
    /// クラス StringReader に単語の読み取り機能を追加したクラスです。<p/>
    /// </summary>
    /// <remarks>
    /// StringReader は、行の読み取りはできますが、単語の読み取りが
    /// できません。この OhStringReader は、
    /// 
    /// <ul>
    /// <li> readWord() によって、空白文字(半角空白、タブ、改行文字)
    /// を区切り文字と見なして、単語を切り出すことができます。</li>
    /// <li> コメントを読み飛ばすように指定できます。指定すれば、行頭に
    /// # がある行は無視されます。行の途中に、# があると、そこか
    /// ら行末まで無視されます。</li>
    /// </ul>
    /// 
    /// コメント行を読み飛ばすように指定する場合は、
    /// <a href ="OhymosNet.OhStringReaderConstructor.html">OhStringReaderのコンストラクタ</a>
    /// のうち、コメントモードを指定できるもので、引数でtrueと指定するか、
    /// オブジェクト生成後に
    /// <a href ="OhymosNet.OhStringReader.TreatCommentP.html">TreatCommentP{get; set;}</a>を
    /// trueに設定します。<p/>
    /// 
    /// この OhStringReader は、データを読み取るために、ストリーム
    /// から1行を取得し、それをメモリに保持しています。メモリ内の文字
    /// 列を走査して、単語を切り出します。<p/>
    /// 
    /// コンストラクタ以外に定義されている関数は、
    /// 
    /// pushBack(), read(), readLine(), readWord(),
    /// skipWhites()
    /// 
    /// です。これ以外に親クラスから継承した関数で使用して良いものは、
    /// close() です。
    /// </remarks>
    public class OhStringReader:StringReader
    {
        
        /// <summary>
        /// 次の行の文字列です。
        /// </summary>
        protected internal String _nextLine;

        /// <summary>
        /// 列のindexを表します。1列目のindexは0です。
        /// </summary>
        protected internal int _ipos;

        /// <summary>
        /// コメントモードにするかどうかのbool値を表します。
        /// 
        /// </summary>
        protected internal bool _treatCommentP = false;

        /// <summary>
        /// 行のindexを表します。1列目のindexは1です。
        /// </summary>
        protected internal int _lineno = 0;

        /// <summary>
        /// 行区切り文字を表します。
        /// </summary>
        protected internal static String NEWLINE = Environment.NewLine;
               
        /// <summary>
        /// テストのためのプロパティです。_nextLineの値を取得します。
        /// </summary>
        public String NextLine
        {
            get { return _nextLine; }
        }
        /// <summary>
        /// テストのためのプロパティです。_iposの値を取得します。
        /// </summary>
        public int Ipos
        {
            get { return _ipos; }
        }
        /// <summary>
        /// テストのためのプロパティです。_linenoの値を取得します。
        /// </summary>
        public int LineNo
        {
            get { return _lineno; }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="str"></param>
        public OhStringReader(string str): base(str)
           { }
        
        /// <summary> Sets _nextLine
        /// 
        /// 
        /// </summary>
        private void setNextLine()
        {
            try
            {
                _nextLine = base.ReadLine();
                _lineno++;
                _ipos = 0;
                while (_nextLine != null && _nextLine.Length > 0 && _nextLine[0] == '#' && _treatCommentP)
                {
                    _nextLine = base.ReadLine();
                    _lineno++;
                }
                if (_nextLine != null && _nextLine.Length > 0)
                {
                    bool incomment = false;
                    char qMark = '\"'; //; "
                    int i = _ipos;
                    for (; i < _nextLine.Length; i++)
                    {
                        char c = _nextLine[i];
                        if (incomment)
                        {
                            if (c == qMark)
                            {
                                incomment = false;
                            }
                            continue;
                        }
                        else
                        {
                            if (c == '\"' || c == '\'')
                            {
                                incomment = true;
                                qMark = c;
                                continue;
                            }
                            else if (_treatCommentP && c == '#')
                            {
                                _nextLine = _nextLine.Substring(0, i - 0);
                                break;
                            }
                        }
                    } // end of for loop
                    if (incomment)
                    {
                        String message = "Input data error. Unballanced quotation marks: " + NEWLINE + _lineno + ": " + _nextLine;
                        // System.err.println(message);
                        throw new IOException(message);
                    }
                } // if (_nextLine != null && _nextLine.length() > 0)
            }
            // try
            catch (IOException e)
            {
                // e.printStackTrace();
                // System.err.println("IO error");
                throw e;
            }
        }

        /// <summary> Push back the last character which had been read in to this
        /// Reader. Only the character which bad been read in from the
        /// current line can be pushed back.
        /// 
        /// <japanese>
        /// <br/>
        /// この Reader に最後に読み込まれた文字を読まなかったことにし
        /// て、元に戻します。ただし、戻すことができるのは、現在読み込
        /// みの対象になっている行から読まれた文字だけです。
        /// </japanese>
        /// </summary>
        public virtual void pushBack()
        {
            if (_lineno == 0)
                return;
            //; まだ1行も読んでいないので、pushBack できないはず。
            if (_ipos > 0)
                _ipos--;
        }

        /// <summary> Read one character. When eof was deteced, return -1.
        /// When the reading pointer is at the line end, return '\n'.
        /// 
        /// </summary>
        /// <returns> an <var>int</var> value
        /// </returns>
        /// <exception cref="IOException">if an error occurs
        /// </exception>
        public override int Read()
        {
            if (_lineno == 0)
                setNextLine();
            if (_nextLine == null)
                return -1;
            if (_nextLine.Length == 0)
            {
                setNextLine();
                return '\n';
            }
            if (_ipos >= _nextLine.Length)
            {
                setNextLine();
                return '\n';
            }
            else
            {
                int k = (char)_nextLine[_ipos];
                _ipos++;
                return k;
            }
        }

        /// <summary> Reads a line of text.
        /// 
        /// <japanese>
        /// <br/>
        /// 現在の読み取り位置から行末まで読み取って返します。<p/>
        /// 
        /// 現在の読みとり位置がファイル終端に来ているときは、
        /// readLine() は、null を返します。<p/>
        /// 
        /// 現在の読み取り位置が行末にあるときは、空き文字列を返します。
        /// <p/>
        /// 
        /// つぎの読み取りポイントは、このメソッドで読取った行のつ
        /// ぎの行の行頭になります。<p/>
        /// </japanese>
        /// 
        /// </summary>
        /// <returns> a <var>String</var> value
        /// </returns>
        /// <exception cref="IOException">if an error occurs
        /// </exception>
        public override String ReadLine()
        {
            if (_lineno == 0)
                setNextLine();
            String lastLine = _nextLine;
            String cline = lastLine;

            if (lastLine != null)
            {
                cline = lastLine.Substring(_ipos, lastLine.Length - _ipos);
                setNextLine();
            }

            else return null;

            return cline;
        }

        /// <summary> Read a word.
        /// 
        /// <japanese>
        /// <br/>
        /// 単語を読みとります。読み取り位置が行末なら、つぎの行から単
        /// 語を読みます。単語を読み取ろうとしたときに、空行 (行頭にす
        /// ぐ改行がある行) に遭遇したら、空き文字列を返します。<p/>
        /// 
        /// 単語とは、
        /// 
        /// <ol>
        /// <li> 引用符号以外の非空白文字の連続</li>
        /// <li> 引用符号 (" または ') から引用符号までの文字列</li>
        /// </ol>
        /// 
        /// を言います。
        /// </japanese>
        /// 
        /// </summary>
        /// <returns> a <var>String</var> value
        /// </returns>
        /// <exception cref="IOException">if an error occurs
        /// </exception>
        public virtual String readWord()
        {
            if (_lineno == 0)
                setNextLine();
            if (_nextLine == null)
                return null;
            int i = 0;
            char c = 'a';
            while (true)
            {
                // search a word
                if (_nextLine.Length == 0)
                {
                    setNextLine();
                    return "";
                }
                if (_ipos == _nextLine.Length)
                {
                    // blank line or now at the line end
                    setNextLine();
                    if (_nextLine == null)
                        return null;
                    continue;
                }
                for (i = _ipos; i < _nextLine.Length; i++)
                {
                    c = _nextLine[i];
                    if (c != ' ' && c != '\t')
                        break;
                }
                if (i < _nextLine.Length)
                {
                    _ipos = i;
                    break; //; find a word
                }
                setNextLine();
            }

            bool quotedStringP = false;
            char qMark = '\"';
            if (c == '\"' || c == '\'')
            {
                quotedStringP = true;
                qMark = c;
                // System.out.println("_ipos = " + _ipos);
            }

            for (i = _ipos + 1; i < _nextLine.Length; i++)
            {
                c = _nextLine[i];
                if (quotedStringP)
                {
                    if (c == qMark)
                    {
                        i++; break;
                    }
                }
                else
                {
                    if (c == ' ' || c == '\t' || c == '\"' || c == '\'')
                        break;
                }
            }
            String word = _nextLine.Substring(_ipos, (i) - (_ipos));
            _ipos = i;
            return word;
        }

        /// <summary> Skip white letters. The line end mark is also skipped.
        /// 
        /// <japanese>
        /// <br/>
        /// 空白文字 (半角空白、タブ文字、改行文字) を読み飛ばし、非空
        /// 白文字の上に読み取りポイントを移動します。
        /// 
        /// 空白文字を読み飛ばしている間に、ファイルの終端に達したとき
        /// は、-1 を返します。それ以外のときは、読み取りポイントの文字
        /// のコードを返します。
        /// </japanese>
        /// </summary>
        public virtual int skipWhites()
        {
            if (_lineno == 0)
                setNextLine();
            if (_nextLine == null)
            {
                return -1;
            }
            int i = 0;
            char c = 'c';
            while (true)
            {
                // search a word
                if (_nextLine.Length == 0 || _ipos == _nextLine.Length)
                {
                    // blank line or now at the line end
                    setNextLine();
                    if (_nextLine == null)
                    {
                        return -1;
                    }
                    continue;
                }
                for (i = _ipos; i < _nextLine.Length; i++)
                {
                    c = _nextLine[i];
                    if (c != ' ' && c != '\t')
                        break;
                }
                if (i < _nextLine.Length)
                {
                    _ipos = i;
                    break; //; find a word
                }
                setNextLine();
            }
            _ipos++;
            return c;
        }

       
    }
}
