﻿// Modified and tested by Naoto Ohshima. Date: 2007/07/05 00:56:22
// Last update:2008/03/03 22:47:23

using System;
using System.Runtime.Serialization;
using System.Text;
using OhymosNet;

namespace OhymosNet
{


    /// <summary> <japanese>
    /// 年と経過時間[sec]とで表されるシミュレーション時刻を表すクラスです。
    /// </japanese>
    /// 
    /// </summary>
    /// 
    /// <remarks>
    /// OhTImePoint は、時刻を表わすクラスです。内部的には、時刻を西
    /// 暦年 <b>year</b>, その年の当初からの経過秒数 <b>tsec</b>、小
    /// 数部分 <b>frac</b> で記録します。<b>frac</b> は、-0.5 以上、
    /// 0.5 未満とします。<p/>
    /// 
    /// 例えば、2003 年の年頭から 10.3 秒経過した時刻を表わす 
    /// OhTimePoint オブジェクトでは、<b>year</b> が 2003, <b>tsec</b> は
    /// 10, <b>frac</b> は、0.3 です。<p/>
    /// <br/>
    /// The OhTimePoint class is the class which expresses time
    /// points. The OhTimePoint class espresses a time point by
    /// <b>year</b>, <b>tsec</b>, and <b>frac</b>. <p/>
    /// 
    /// <b>tsec</b> is the time measured in seconds from the start
    /// of the year. <b>frac</b> is the fraction part of the time
    /// which is greater than or equal to -0.5 and less than 0.5.<p/>
    /// </remarks>
    public class OhTimePoint : ICloneable, IComparable,IOhTimePointImmutable //ISerializable,
    {
        //@begin{fields}---------------------------------------------------------------------

        /// <summary> <japanese>
        /// 年です。
        /// </japanese>
        /// 
        /// </summary>
        protected internal int _year;

        /// <summary> <japanese>
        /// 対象とする年の年始からの経過時間[sec]です。
        /// </japanese>
        /// 
        /// </summary>
        protected internal long _tsec;

        /// <summary>
        /// 経過時間[sec]の小数部分です。-0.5以上、0.5未満とします。
        /// </summary>
        protected internal double _frac;

        /// <summary>
        /// 対象とする年の年間の総秒数[sec]です
        /// </summary>
        protected internal long _oneYearInSec;
        /// <summary>
        /// うるう年でない年の月の日数の配列です。
        /// 要素は1月から12月の日数です。
        /// </summary>
        protected internal static int[] nDay = new int[] { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

        /// <summary>
        /// うるう年の月の日数の配列です。
        /// 要素は1月から12月の日数です。
        /// </summary>
        protected internal static int[] lDay = new int[] { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

        /// <summary> 
        /// 西暦年を表す配列のインデックス（定数）です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of year. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int YEAR = 0;
        /// <summary> 
        /// 月を表す定数です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of month. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int MONTH = 1;
        /// <summary> 
        /// 日を表す配列のインデックス（定数）です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of day. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int DAY = 2;
        /// <summary> 
        /// 時間を表す配列のインデックス（定数）です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of hour. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int HOUR = 3;
        /// <summary> 
        /// 分を表す配列のインデックス（定数）です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of munute. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int MINUTE = 4;
        /// <summary> 
        /// 秒を表す配列のインデックス（定数）です。
        /// <a href = "OhymosNet.OhTimePoint.toCalendarArray.html">toCalendarArray()</a>で
        /// 使用されます。<p/>
        /// The index of the element of the calendar array which contains the
        /// datum of second. The calendar array is obtained by toCalendarArray().
        /// </summary>
        public const int SECOND = 5;

        /// <summary>
        /// 現実的にこれ以上昔の時刻はないと思われる時刻です。<p/>
        /// <ul>
        /// <li> _year : -1</li>
        /// <li> _tsec : 0</li>
        /// </ul>
        /// <p/>
        /// をもって定義しています。
        /// 
        /// </summary>
        public static readonly OhTimePoint OhTimePointMin = new OhTimePoint(-1, 0L);

        /// <summary>
        /// 現実的にこれ以上最新の時刻はないと思われる時刻です。<p/>
        /// <ul>
        /// <li> _year : 10000</li>
        /// <li> _tsec : 0</li>
        /// </ul>
        /// <p/>
        /// をもって定義しています。
        /// 
        /// </summary>
        public static  readonly OhTimePoint OhTimePointMax = new OhTimePoint(10000, 0L);

        /// <summary> 
        /// 西暦0年0時0分0秒0.0を持つOhTimePointオブジェクトです。<p/>
        /// OhTimePoint object with <b>year</b> = 0, <b>sec</b> = 0 and
        /// <b>frac</b> = 0.0.
        /// 
        /// </summary>
        public static readonly OhTimePoint ZERO = new OhTimePoint(0, 0L, 0.0);

        //@end{fields}--------------------------------------------------------------------
        
        //@begin{constructers}------------------------------------------------------------
        
        /// <summary> 
        /// OhTimePointの新しいオブジェクトを生成します。
        /// デフォルトコンストラクタです。<p/>
        /// Creates a new default <var>OhTimePoint</var> instance.
        /// The default time is the beginning of the year 0.
        /// </summary>
        /// <remarks>
        /// メンバ変数は次のように初期化されます。
        /// <ul>
        /// <li>_year = 0</li>
        /// <li>_tsec = 0</li>
        /// <li>_frac = 0</li>
        /// <li>_oneYearInSec = 31622400（うるう年の総秒数）</li>
        /// </ul>
        /// </remarks>
        /// 
        public OhTimePoint()
        {
            _year = 0; _tsec = 0; _frac = 0; _oneYearInSec = 31622400;
        }

        /// <summary> 
        ///  aYear 年の初頭から aTsec 秒の時刻の OhTimePoint オ
        /// ブジェクトを生成します。<p/>
        /// 
        /// Creates a new <var>OhTimePoint</var> instance.
        /// 
        /// </summary>
        /// <param name="aYear">an <var>int</var> value<p/>
        /// 西暦年
        /// </param>
        /// <param name="aTsec">a <var>long</var> value<p/>
        /// 年始からの秒数<p/>
        /// Time from the beginning
        /// of the year.
        /// </param>
        /// <exception cref="ArgumentException">
        /// aTsecが0より小さい、または_oneYearInSecより大きい場合
        /// </exception>
        public OhTimePoint(int aYear, long aTsec)
        {
            _year = aYear;
            if (_year == 0)
            {
                _oneYearInSec = 31622400;
            }
            // 366 * 24 * 3600
            else
            {
                _oneYearInSec = 31536000; // 365 * 24 * 3600
                if (isLeapYear(_year))
                    _oneYearInSec = 31622400;
            }
            if (aTsec < 0 || aTsec >= _oneYearInSec)
            {
                throw new ArgumentException("illegal aTsec");
            }
            _tsec = aTsec;
            _frac = 0;
        }

        /// <summary> 
        /// aYear 年の初頭から aTsec + aFrac 秒の時刻の OhTimePoint オ
        /// ブジェクトを生成します。
        /// <p/>
        /// 
        /// Creates a new <var>OhTimePoint</var> instance.
        /// <remarks>
        /// このとき、引数aFracは-0.5以上、0.5未満でなければなりません。<p/>
        /// aFrac must be greater than or equal to -0.5 and less than 0.5.
        /// 
        /// </remarks>
        /// 
        /// </summary>
        /// <param name="aYear">an <var>int</var> value<p/>
        /// 西暦年
        /// </param>
        /// <param name="aTsec">a <var>long</var> value<p/> 
        /// 年始からの秒数の整数部分
        /// Time in seconds from the beginning.
        /// </param>
        /// <param name="aFrac">an <var>double</var> value<p/> 
        /// 年始からの秒数の端数部分
        /// </param>
        /// <exception cref="ArgumentException">
        /// aTsecが0より小さい、または_oneYearInSecより大きい場合
        /// </exception>
        public OhTimePoint(int aYear, long aTsec, double aFrac)
        {
            _year = aYear;
            if (_year == 0)
            {
                _oneYearInSec = 31622400;
            }
            // 366 * 24 * 3600
            else
            {
                _oneYearInSec = 31536000; // 365 * 24 * 3600
                if (isLeapYear(_year))
                    _oneYearInSec = 31622400;
            }
            if (aTsec < 0 || aTsec >= _oneYearInSec)
            {
                throw new ArgumentException("illegal aTsec");
            }
            _tsec = aTsec;
            _frac = aFrac;
        }

        /// <summary> 
        /// aYear 年 aMonth 月 aDay 日 aHour 時 aMin 分 aSec 秒
        /// に対応するOhTimePoint オブジェクトを生成します。
        /// <p/>
        /// Creates a new <var>OhTimePoint</var> instance.
        /// </summary>
        /// <remarks>
        /// 引数は以下の範囲で指定します。<p/>
        /// <ul>
        /// <li>aYear   0-</li>
        /// <li>aMonth  1-12</li>
        /// <li>aDay    1-31</li>
        /// <li>aHour   0-23</li>
        /// <li>aMin    0-59</li>
        /// <li>aSec    0-59</li>
        /// </ul>
        /// これらの引数を用いてメンバ変数は以下のように初期化されます。
        /// <ul>
        /// <li>_year = aYear</li>
        /// <li>_tsec = 引数より計算された総秒数</li>
        /// <li>_frac = 0    </li>
        /// <li>_oneYearSec = 31536000（通常年） or 31622400（うるう年）</li>
        /// </ul>
        /// </remarks>
        /// <param name="aYear">an <var>int</var> value<p/> 
        /// 西暦年
        /// </param>
        /// <param name="aMonth">an <var>int</var> value<p/> 
        /// 月
        /// </param>
        /// <param name="aDay">an <var>int</var> value<p/> 
        /// 日
        /// </param>
        /// <param name="aHour">an <var>int</var> value<p/> 
        /// 時
        /// </param>
        /// <param name="aMin">an <var>int</var> value<p/> 
        /// 分
        /// </param>
        /// <param name="aSec">an <var>int</var> value<p/>
        /// 秒
        /// </param>
        /// <exception cref="ArgumentException">
        /// 引数の範囲が間違っている場合
        /// </exception>
        public OhTimePoint(int aYear, int aMonth, int aDay, int aHour, int aMin, int aSec)
        {
            _year = aYear;
            bool wLeapYearP = isLeapYear(aYear);
            _oneYearInSec = 31536000; // 365 * 24 * 3600
            if (wLeapYearP)
                _oneYearInSec += 86400;

            _tsec = 0;

            if (aMonth < 1 || aMonth > 12)
            {
                throw new ArgumentException("illegal aMonth");
            }
            for (int i = 0; i < aMonth - 1; i++)
            {
                _tsec += (wLeapYearP ? lDay[i] : nDay[i]) * 86400;
            }
            if (aDay < 1 || aDay > (wLeapYearP ? lDay[aMonth - 1] : nDay[aMonth - 1]))
            {
                throw new ArgumentException("illegal aDay");
            }
            _tsec += (aDay - 1) * 86400;
            if (aHour < 0 || aHour >= 24)
            {
                throw new ArgumentException("illegal aHour");
            }
            _tsec += aHour * 3600;
            if (aMin < 0 || aMin >= 60)
            {
                throw new ArgumentException("illegal aMin");
            }
            _tsec += aMin * 60;
            if (aSec < 0 || aSec >= 60)
            {
                throw new ArgumentException("illegal aSec");
            }
            _tsec += aSec;
            _frac = 0;
        }

        /// <summary> 
        /// aYear 年 aMonth 月 aDay 日 aHour 時 aMin 分 aSec + aFrac 秒
        /// に対応するOhTimePoint オブジェクトを生成します。
        /// <p/>
        /// Creates a new <var>OhTimePoint</var> instance.
        /// 
        /// </summary>
        /// <remarks>
        /// 引数は以下の範囲で指定します。<p/>
        /// <ul>
        /// <li>aYear   0-</li>
        /// <li>aMonth  1-12</li>
        /// <li>aDay    1-31</li>
        /// <li>aHour   0-23</li>
        /// <li>aMin    0-59</li>
        /// <li>aSec    0-59</li>
        /// <li>aFrac   -0.5以上0.5未満</li>
        /// </ul>
        /// これらの引数を用いてメンバ変数は以下のように初期化されます。
        /// <ul>
        /// <li>_year = aYear</li>
        /// <li>_tsec = 引数より計算された総秒数</li>
        /// <li>_frac = aFrac</li>
        /// <li>_oneYearSec = 31536000（通常年） or 31622400（うるう年）</li>
        /// </ul>
        /// </remarks>
        /// <param name="aYear">an <var>int</var> value<p/>
        /// 西暦年
        /// </param>
        /// <param name="aMonth">an <var>int</var> value<p/>
        /// 月
        /// </param>
        /// <param name="aDay">an <var>int</var> value<p/> 
        /// 日
        /// </param>
        /// <param name="aHour">an <var>int</var> value<p/>
        /// 時
        /// </param>
        /// <param name="aMin">an <var>int</var> value<p/>
        /// 分
        /// </param>
        /// <param name="aSec">an <var>int</var> value<p/>
        /// 秒
        /// </param>
        /// <param name="aFrac">an <var>double</var> value<p/>
        /// </param>
        /// <exception cref="ArgumentException">
        /// 引数の範囲が間違っている場合
        /// </exception>
        public OhTimePoint(int aYear, int aMonth, int aDay, int aHour, int aMin, int aSec, double aFrac)
        {
            _year = aYear;
            bool wLeapYearP = isLeapYear(aYear);
            _oneYearInSec = 31536000; // 365 * 24 * 3600
            if (wLeapYearP)
                _oneYearInSec += 86400;

            _tsec = 0;

            if (aMonth < 1 || aMonth > 12)
            {
                throw new ArgumentException("illegal aMonth");
            }
            for (int i = 0; i < aMonth - 1; i++)
            {
                _tsec += (wLeapYearP ? lDay[i] : nDay[i]) * 86400;
            }
            if (aDay < 1 || aDay > (wLeapYearP ? lDay[aMonth - 1] : nDay[aMonth - 1]))
            {
                throw new ArgumentException("illegal aDay");
            }
            _tsec += (aDay - 1) * 86400;
            if (aHour < 0 || aHour >= 24)
            {
                throw new ArgumentException("illegal aHour");
            }
            _tsec += aHour * 3600;
            if (aMin < 0 || aMin >= 60)
            {
                throw new ArgumentException("illegal aMin");
            }
            _tsec += aMin * 60;
            if (aSec < 0 || aSec >= 60)
            {
                throw new ArgumentException("illegal aSec");
            }
            _tsec += aSec;
            _frac = aFrac;
        }

        /// <summary> 
        /// 他の<a href = "OhymosNet.IOhTimePointImmutable.html">
        /// IOhTimePointImmutable</a>オブジェクトからOhTimePointの
        /// 新しいオブジェクトを生成します。<p/>
        /// Creates a new <var>OhTimePoint</var> instance from another
        /// <var>OhTimePoint</var> instance.
        /// 
        /// </summary>
        /// <param name="tp">an <var>IOhTimePointImmutable</var> value<p/>
        /// </param>
        public OhTimePoint(IOhTimePointImmutable tp)
        {
            _year = tp.Year;
            _tsec = tp.Tsec;
            _frac = tp.Frac;
            _oneYearInSec = tp.OneYearInSec;
        }

        //@end{constructers}--------------------------------------------------------------
        
        //@begin{properties}----------------------------------------------------------------

        /// <summary> 
        /// 西暦年を取得、または
        /// 西暦年、その年の秒数を設定します。<p/>
        /// Gets and the year, and sets the year and the seconds from the beginning of it.
        /// 
        /// </summary>
        /// <exception cref="ArgumentException">
        /// 設定した値が-1より小さい場合<p/>
        /// </exception>
        public int Year
        {
            get { return _year; }
            set
            {
                if (value < -1)
                    throw new ArgumentException("illegal value");

                _year = value;
                _oneYearInSec = 365 * 24 * 3600;

                if (isLeapYear(_year))
                    _oneYearInSec += 86400;
                _tsec = 0;
                _frac = 0;
            }
        }


        /// <summary> 
        /// 年始からの秒数を取得・設定します。<p/>
        /// Gets the time from the beginning of the year of the time expressed
        /// by this OhTimePoint object.<p/>
        /// Sets this OhTimePoint time at the time point when 'aTsec' seconds
        /// have elapsed from the beginning of the year.
        /// 
        /// </summary>
        /// <exception cref="ArgumentException">
        /// 設定する値（経過秒数）が0未満、または_oneYearInSecよりも
        /// 大きいとき
        /// </exception>
        public long Tsec
        {
            get { return _tsec; }
            set
            {
                if (value < 0 || value > _oneYearInSec)
                    throw new ArgumentException("illegal value");

                _tsec = value;
            }
        }

        /// <summary>
        /// 経過秒数の小数部分を取得・設定します。<p/>
        /// Gets and sets the fraction of seconds.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// 設定した値（小数部分）が-0.5未満、または0.5以上のとき
        /// </exception>
        public double Frac
        {
            get { return _frac; }
            set
            {
                if (value < -0.5 || value >= 0.5)
                    throw new ArgumentException("illegal value");

                _frac = value;
            }
        }

        /// <summary> 
        /// その年の総秒数を取得します。<p/>
        /// Gets the number of seconds contained in the year of the time expressed
        /// by this OhTimePoint object
        /// 
        /// </summary>
        public long OneYearInSec
        {
            get { return _oneYearInSec; }
        }
        //@end{properties}------------------------------------------------------------------

        //@begin{methods}----------------------------------------------------------------

        /// <summary> 
        /// Object型にキャストされたOhTimePointオブジェクトのディープコピーを返します。<p/>
        /// Returns a deep copy of the current object
        /// 
        /// </summary>
        /// <returns> an <var>Object</var> value<p/>
        /// </returns>
        /// 
        public Object Clone()
        {
            return new OhTimePoint(_year, _tsec, _frac);
        }

        /// <summary> 
        /// 引数tpのメンバ変数をコピーします。<p/>
        /// Copies fields from those of <var>OhTimePoint</var> tp.
        /// 
        /// </summary>
        /// <param name="tp">an <var>Object</var> value<p/>
        /// コピー元の<a href = "OhymosNet.IOhTimePointImmutable.html">
        /// IOhTimePointImmutable</a>オブジェクト
        /// </param>
        public virtual void copy(IOhTimePointImmutable tp)
        {
            _year = tp.Year;
            _tsec = tp.Tsec;
            _frac = tp.Frac;
            _oneYearInSec = tp.OneYearInSec;
        }

        /// <summary> 
        /// 引数に指定したオブジェクトが OhTimePointImmutable オブジェ
        /// クトであって、かつ、現在の OhTimePointImmutable オブジェク
        /// トと同じ状態であるかをテストします。<p/>
        /// Tests if the specified Object is an instance of
        /// OhTimePointImmutable and if it is equal to this
        /// OhTimePoint.
        /// </summary>
        /// <param name="arg">an <var>Object</var> value<p/>
        /// 比較するオブジェクト
        /// </param>
        /// <returns> a <var>boolean</var> value<p/>
        /// </returns>
        public override bool Equals(Object arg)
        {
            return ((arg != null) && (arg is IOhTimePointImmutable) && _year == ((OhTimePoint)arg)._year 
                && _tsec == ((OhTimePoint)arg)._tsec && _frac==((OhTimePoint)arg)._frac);
        }

        /// <summary> 
        /// 引数に指定したオブジェクト <b>anObj</b> が 
        /// OhTimePointImmutable オブジェクトであって、かつ、現在の 
        /// OhTimePointImmutable オブジェクトとの時間差が <b>anEps</b> 
        /// より小さいかをテストします。<p/>
        /// Tests if the specified Object is an instance of
        /// OhTimePointImmutable and if the time differenc between the
        /// current time point and <b>anObj</b> is less than <b>anEps</b>.
        /// </summary>
        /// <remarks>
        /// <b>anEps</b> の値として、
        /// <a href ="OhymosNet.OhTime.NEGLIGIBLY_SMALL_TIME.html">
        /// OhTime.NEGLIGIBLY_SMALL_TIME.html</a>を使用するとよいでしょう。
        /// </remarks>
        /// 
        /// <param name="anObj">an <var>Object</var> value<p/>
        /// 比較するオブジェクト
        /// </param>
        /// <param name="anEps">a <var>double</var> value<p/>
        /// 指定する時間差
        /// </param>
        /// <returns> a <var>boolean</var> value<p/>
        /// </returns>
        public bool equals(Object anObj, double anEps)
        {
            return ((anObj != null) && (anObj is IOhTimePointImmutable) && _year == ((OhTimePoint)anObj)._year && _tsec == ((OhTimePoint)anObj)._tsec && Math.Abs(((OhTimePoint)anObj)._frac - _frac) < anEps);
        }
       
        /// <summary> 
        /// 引数aYearがうるう年かを判定します。
        /// <p/>
        /// Tests if the year 'aYear' is a leap year or not.
        /// 
        /// </summary>
        /// <param name="aYear">an <var>int</var> value<p/>
        /// 西暦年
        /// </param>
        /// <returns> a <var>boolean</var> value<p/>
        /// <ul><pre>
        /// <li>true    うるう年である</li>
        /// <li>false   うるう年でない</li>
        /// </pre></ul>
        /// </returns>
        public static bool isLeapYear(int aYear)
        {
            if (aYear % 4 != 0)
                return false;
            if (aYear % 100 != 0)
                return true;
            if (aYear % 400 != 0)
                return false;
            return true;
        }

        /// <summary> 
        /// 時刻の配列（{年、月、日、時、分、秒}）を返します。<p/>
        /// Converts the current OhTimePoint object to the Calendar array.
        /// </summary>
        ///  
        /// <remarks>
        /// When cal is the returned array, cal[OhTimePoint.YEAR],
        /// cal[OhTimePoint.MONTH], cal[OhTimePoint.DAY],
        /// cal[OhTimePoint.HOUR], cal[OhTimePoint.MINUTE],
        /// cal[OhTimePoint.SECOND] contain year, month, hour, minute and
        /// second data.
        /// 
        /// </remarks>
        /// <returns> an <var>int[]</var> value<p/>
        /// </returns>
        public virtual int[] toCalendarArray()
        {
            int[] cal = new int[6];

            int year = _year;
            int month = 1;
            int day = 1;
            int hour = 0;
            int minute = 0;
            int second = 0;
            int wT = (int)_tsec;
            int yearSec;
            bool wLeapYearP;
            while (true)
            {
                wLeapYearP = isLeapYear(year);
                yearSec = 31536000; // 365 * 24 * 3600;
                if (wLeapYearP)
                {
                    yearSec += 86400;
                }
                if (wT <= yearSec)
                    break;
                year += 1;
                wT -= yearSec;
            }
            long monthSec;
            while (true)
            {
                monthSec = (wLeapYearP ? lDay[month - 1] : nDay[month - 1]) * 86400;
                if (wT <= monthSec)
                    break;
                month += 1;
                wT = (int)(wT - monthSec);
            }
            day = wT / 86400 + 1;
            wT -= (day - 1) * 86400;
            hour = wT / 3600;
            wT -= hour * 3600;
            minute = wT / 60;
            wT -= minute * 60;
            second = wT;

            cal[YEAR] = year; cal[MONTH] = month; cal[DAY] = day;
            cal[HOUR] = hour; cal[MINUTE] = minute; cal[SECOND] = second;

            return cal;
        }

       

        /// <summary> Compares the current OhTimePoint object to another 
        /// object 'arg' and returns the result.
        /// The returend value is 1, 0, -1 
        /// </summary>
        /// <param name="arg">an <var>Object</var> value<p/>
        /// </param>
        /// <returns> an <var>int</var> value<p/>
        /// the value 0 if this OhTimePoint object expresses the same time as
        /// the argument OhTimePoint object; 
        /// the value -1 0 if this OhTimePoint object expresses earlier
        /// time than the
        /// argument OhTimePoint object;
        /// the value 1 if this OhTimePoint object expresses later time than the
        /// argument OhTimePoint object.
        /// </returns>
        public virtual int CompareTo(Object arg)
        {
            OhTimePoint a = (OhTimePoint)arg;
            if (_year > a._year)
                return 1;
            if (_year < a._year)
                return -1;
            if (_tsec > a._tsec)
                return 1;
            if (_tsec < a._tsec)
                return -1;
            if (_frac > a._frac)
                return 1;
            if (_frac < a._frac)
                return -1;
            return 0;
        }
        /// <summary> Compares the current OhTimePoint object to another 
        /// object 'arg' and returns the result.
        /// The returend value is 1, 0, -1 
        /// </summary>
        /// <param name="arg">an <var>Object</var> value<p/>
        /// </param>
        /// <returns> an <var>int</var> value<p/>
        /// the value 0 if this OhTimePoint object expresses the same time as
        /// the argument OhTimePoint object; 
        /// the value -1 0 if this OhTimePoint object expresses earlier
        /// time than the
        /// argument OhTimePoint object;
        /// the value 1 if this OhTimePoint object expresses later time than the
        /// argument OhTimePoint object.
        /// </returns>
        public int CompareTo(IOhTimePointImmutable arg)
        {
            if (_year > arg.Year)
                return 1;
            if (_year < arg.Year)
                return -1;
            if (_tsec > arg.Tsec)
                return 1;
            if (_tsec < arg.Tsec)
                return -1;
            if (_frac > arg.Frac)
                return 1;
            if (_frac < arg.Frac)
                return -1;
            return 0;
        }

        /// <summary> Adds the specified amount of seconds to the current object and
        /// returns the updated object.
        /// 
        /// </summary>
        /// <param name="timeSec">a <var>long</var> value<p/> 
        /// </param>
        /// <returns> an <var>OhTimePoint</var> value<p/>
        /// </returns>
        public void add(long timeSec)
        {
            if (timeSec >= 0)
            {
                _tsec += timeSec;
                while (_tsec >= _oneYearInSec)
                {
                    _tsec -= _oneYearInSec;
                    _year += 1;
                    _oneYearInSec = 31536000L; // 365 * 24 * 3600;
                    if (isLeapYear(_year))
                        _oneYearInSec = 31622400L;
                }
            }
            else
            {
                _tsec += timeSec;
                while (_tsec < 0)
                {
                    _year -= 1;
                    _oneYearInSec = 31536000L; // 365 * 24 * 3600;
                    if (isLeapYear(_year))
                        _oneYearInSec = 31622400L;
                    _tsec += _oneYearInSec;
                }
            }
        }

        /// <summary> Adds the specified amount of seconds to the current object.
        /// <p/>
        /// 
        /// <japanese>
        /// <p/>
        /// 現在の OhTimePoint オブジェクトに OhTimeImmutable time で表
        /// わされる時間を足します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="time">a <var>IOhTimeImmutable</var> value<p/> 
        /// </param>
        public void add(IOhTimeImmutable time)
        {
            double frac = _frac + time.Frac;
            long n = (long)Math.Round(frac);
            _frac = frac - n;
            long timeSec = time.Sec + n;
            if (timeSec >= 0)
            {
                _tsec += timeSec;
                while (_tsec >= _oneYearInSec)
                {
                    _tsec -= _oneYearInSec;
                    _year += 1;
                    _oneYearInSec = 31536000L; // 365 * 24 * 3600;
                    if (isLeapYear(_year))
                        _oneYearInSec = 31622400L;
                }
            }
            else
            {
                _tsec += timeSec;
                while (_tsec < 0)
                {
                    _year -= 1;
                    _oneYearInSec = 31536000L; // 365 * 24 * 3600;
                    if (isLeapYear(_year))
                        _oneYearInSec = 31622400L;
                    _tsec += _oneYearInSec;
                }
            }
        }

        /// <summary> Adds <b>aTime</b> to <b>aTp</b> and returns the resultant
        /// object.
        /// 
        /// <japanese>
        /// <p/>
        /// 時刻 aTp に、時間 aTime を足した時刻を返します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aTp">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <param name="aTime">an <var>OhTimeImmutable</var> value<p/>
        /// </param>
        /// <returns> an <var>OhTimePoint</var> value<p/>
        /// </returns>
        public static OhTimePoint add(IOhTimePointImmutable aTp, IOhTimeImmutable aTime)
        {
            OhTimePoint tp = new OhTimePoint(aTp.Year, aTp.Tsec, aTp.Frac);
            tp.add(aTime);
            return tp;
        }

        /// <summary> Sutracts aLongTime from the current OhTimePoint object and
        /// returnts the updated object.
        /// 
        /// <japanese>
        /// <p/>
        /// 現在の OhTimePoint オブジェクトから aLongTime 引数で指定す
        /// る時間を引きます。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aLongTime">a <var>long</var> value<p/>
        /// </param>
        public void sub(long aLongTime)
        {
            add(-aLongTime);
        }

        /// <summary> Subtracts aTime from the current OhTimePoint object.
        /// 
        /// <japanese>
        /// <p/>
        /// 現在の OhTimePoint オブジェクトから aTime 引数で指定する時
        /// 間を引きます。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aTime">an <var>OhTimeImmutable</var> value<p/>
        /// </param>
        public void sub(IOhTimeImmutable aTime)
        {
            OhTime t = new OhTime(-aTime.Sec, -aTime.Frac);
            t.adjust();
            add(t);
        }

        /// <summary> Subtract <b>aTime</b> from <b>aTp</b> and returns the
        /// resultant OhTimePoint object.
        /// 
        /// <japanese>
        /// <p/>
        /// 時刻 aTp から、時間 aTime を引いた時刻を返します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aTp">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <param name="aTime">an <var>OhTimeImmutable</var> value<p/>
        /// </param>
        /// <returns> an <var>OhTimePoint</var> value<p/>
        /// </returns>
        public static OhTimePoint sub(IOhTimePointImmutable aTp, IOhTimeImmutable aTime)
        {
            OhTimePoint tp = new OhTimePoint(aTp.Year, aTp.Tsec, aTp.Frac);
            tp.sub(aTime);
            return tp;
        }

       /// <summary> Returns the time difference between the two
        /// OhTimePointImmutable objects.<p/>
        /// 
        /// <japanese>
        /// <p/>
        /// 引数で与える2つの OhTimePointImmutable オブジェクトの時間差
        /// を返します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aT">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <param name="bT">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <returns> a <var>OhTime</var> value<p/>
        /// </returns>
        public static OhTime sub(IOhTimePointImmutable aT, IOhTimePointImmutable bT)
        {
            long t = 0L;
            int yeara = aT.Year;
            long tseca = aT.Tsec;
            int yearb = bT.Year;
            long tsecb = bT.Tsec;

            if (yeara == yearb)
            {
                t = Math.Abs(tseca - tsecb);
            }
            else if (yeara < yearb)
            {
                int y = yeara;
                while (y < yearb)
                {
                    if (isLeapYear(y))
                        t = t + 31622400L;
                    else
                        t = t + 31536000L;
                    y++;
                }
                t = t - tseca + tsecb;
            }
            else
            {
                int y = yearb;
                while (y < yeara)
                {
                    if (isLeapYear(y))
                        t = t + 31622400L;
                    else
                        t = t + 31536000L;
                    y++;
                }
                t = t - tsecb + tseca;
            }
            double frac = aT.Frac - bT.Frac;
            OhTime time = OhTime.adjust(t, frac);
            return time;
        }

        /// <summary> Returns the time difference between the current OhTimePoint
        /// and the OhTimePointImmutable <b>aT</b>.
        /// 
        /// <japanese>
        /// <p/>
        /// 現在の OhTimePoint オブジェクトと、引数で指定した OhTimePoint 
        /// オブジェクト aT の時間差を計算して戻します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aT">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <returns> an <var>OhTime</var> value<p/>
        /// </returns>
        public OhTime sub(IOhTimePointImmutable aT)
        {
            long t = 0L;
            int year = aT.Year;
            long tsec = aT.Tsec;

            if (_year == year)
            {
                t = Math.Abs(_tsec - tsec);
            }
            else if (_year < year)
            {
                int y = _year;
                while (y < year)
                {
                    if (isLeapYear(y))
                        t = t + 31622400L;
                    else
                        t = t + 31536000L;
                    y++;
                }
                t = t - _tsec + tsec;
            }
            else
            {
                int y = year;
                while (y < _year)
                {
                    if (isLeapYear(y))
                        t = t + 31622400L;
                    else
                        t = t + 31536000L;
                    y++;
                }
                t = t - tsec + _tsec;
            }
            double frac = _frac - aT.Frac;
            OhTime time = OhTime.adjust(t, frac);
            return time;
        }

        /// <summary> Compares two OhTimePointImmutable objects <b>aTpa</b> and
        /// <b>aTpb</b>. When the time difference of the two objects
        /// is less than <b>aEps</b>, then 0 is returned.
        /// 
        /// <japanese>
        /// <p/>
        /// 二つの時刻 <b>aTpa</b> と <b>aTpb</b> の時間差を調べます。
        /// 時間差の絶対値が <b>aEps</b> より小さい場合は、実質的に等し
        /// いと見なして、0 を返します。<b>aTpa</b> が
        /// <b>aTpb - aEps</b> より前のときは、-1 を返します。
        /// <b>aTpa</b> が <b>aTpb + aEps</b> より後のときは、
        /// +1 を返します。
        /// </japanese>
        /// 
        /// </summary>
        /// <param name="aTpa">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <param name="aTpb">an <var>OhTimePointImmutable</var> value<p/>
        /// </param>
        /// <param name="aEps">a <var>double</var> value<p/>
        /// </param>
        /// <returns> an <var>int</var> value<p/>
        /// </returns>
        public static int compare(IOhTimePointImmutable aTpa, IOhTimePointImmutable aTpb, double aEps)
        {
            OhTimePoint ta = (OhTimePoint)aTpa.Clone();
            OhTime d = ta.sub(aTpb);
            double diff = d.toDouble();
            if (diff > aEps)
                return 1;
            if (diff < -aEps)
                return -1;
            return 0;
        }

        /// <summary> Returns "" + _year + " " + _tsec + " " + _frac.
        /// 
        /// </summary>
        /// <returns> a <var>String</var> value<p/>
        /// </returns>
        public override String ToString()
        {
            StringBuilder a = new StringBuilder("");
            a.Append(_year).Append(" ").Append(_tsec);
            a.Append(" ").Append(_frac);
            return a.ToString();
        }
	

        /// <summary> Returns the CalendaString. For example, this method returns
        /// "20030101010101 0.0" for 2003 January 1st, 01:01:01 0.0.
        /// 
        /// </summary>
        /// <returns> a <var>String</var> value<p/>
        /// </returns>

        public virtual String toCalendarString()
        {
            int[] cal = toCalendarArray();
            DateTime dt = new DateTime(cal[0], cal[1], cal[2], cal[3], cal[4], cal[5]);

            return dt.ToString("yyyyMMddHHmmss") + " " + _frac.ToString();
        }



        // 以下、何のための機能かわからないのでとりあえず置いておく

        ///// <summary> Writes the state of the current object into aOut. The
        ///// readObject method restore the object by reading the result.
        ///// 
        ///// <japanese>
        ///// <p/>
        ///// aOut に、現在の OhTime オブジェクトの状態を書き出します。
        ///// readObject が、後で、それを読み取ってオブジェクトの状態を復
        ///// 元します。
        ///// </japanese>
        ///// 
        ///// </summary>
        ///// <param name="aOut">an <var>ObjectOutputStream</var> value<p/>
        ///// </param>
        ///// <exception cref="IOException">if an error occurs
        ///// </exception>
        ////UPGRADE_TODO: Class 'java.io.ObjectOutputStream' は、動作の異なる 'System.IO.BinaryWriter' に変換されました。 "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1073_javaioObjectOutputStream'"
        ////UPGRADE_TODO: メソッド 'writeObject' は 'GetObjectData' に変換され、パラメータが変更されました。コードを見直して、変換されたパラメータの非メンバ メソッドに対する呼び出しがないことを確認してください。 "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1192'"
        //public void GetObjectData(SerializationInfo aOut, StreamingContext context)
        //{
        //    SupportClass.DefaultWriteObject(aOut, context, this);
        //    aOut.AddValue("ohymosj.OhTimePointdata1", _year);
        //    aOut.AddValue("ohymosj.OhTimePointdata2", _tsec);
        //    aOut.AddValue("ohymosj.OhTimePointdata3", _frac);
        //    aOut.AddValue("ohymosj.OhTimePointdata4", _oneYearInSec);
        //}

        ///// <summary> Restores the state of the object from aIn.
        ///// 
        ///// <japanese>
        ///// <p/>
        ///// aIn から、OhTime オブジェクトの状態を読み取ります。
        ///// </japanese>
        ///// 
        ///// </summary>
        ///// <param name="aIn">an <var>ObjectInputStream</var> value<p/>
        ///// </param>
        ///// <exception cref="IOException">if an error occurs
        ///// </exception>
        ///// <exception cref="ClassNotFoundException">if an error occurs
        ///// </exception>
        ////UPGRADE_TODO: Class 'java.io.ObjectInputStream' は、動作の異なる 'System.IO.BinaryReader' に変換されました。 "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1073_javaioObjectInputStream'"
        ////UPGRADE_TODO: メソッド 'readObject' は 'OhTimePoint' に変換され、パラメータが変更されました。コードを見直して、変換されたパラメータの非メンバ メソッドに対する呼び出しがないことを確認してください。 "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1192'"
        //private OhTimePoint(SerializationInfo aIn, StreamingContext context)
        //{
        //    SupportClass.DefaultReadObject(aIn, context, this);
        //    _year = aIn.GetInt32("ohymosj.OhTimePointdata1");
        //    _tsec = aIn.GetInt64("ohymosj.OhTimePointdata2");
        //    _frac = aIn.GetDouble("ohymosj.OhTimePointdata3");
        //    _oneYearInSec = aIn.GetInt64("ohymosj.OhTimePointdata4");
        //}

        ///// <summary>
        ///// System.Objectから継承されたメソッド。
        ///// </summary>
        ///// <returns>a <var>int</var> value</returns>
        //public override int GetHashCode()
        //{
        //    return base.GetHashCode();
        //}
        //@end{methods}----------------------------------------------------------------

    } // ohtimepoint
}